@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 2008, 2010, 2012, 2016 Sergey Poznyakoff
+   Copyright (C) 2008, 2010, 2012, 2016-2018 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -41,9 +41,17 @@ int utf8_wctomb (char *r, unsigned int wc);
 
 int utf8_symcmp(char *a, char *b);
 int utf8_symcasecmp(char *a, char *b);
+
+int utf8_strcmp_cc(char *a, char *b, int ci);
+
+int utf8_strcmp(char *a, char *b);
 int utf8_strcasecmp(char *a, char *b);
 int utf8_strncasecmp(char *a, char *b, size_t maxlen);
 
+int utf8_strcmp_alnumspace_cc(char *a, char *b, int ci);
+int utf8_strcmp_alnumspace(char *a, char *b);
+int utf8_strcasecmp_alnumspace(char *a, char *b);
+
 unsigned utf8_wc_toupper (unsigned wc);
 int utf8_toupper (char *s);
 unsigned utf8_wc_tolower (unsigned wc);
@@ -72,6 +80,12 @@ const unsigned *utf8_wc_strstr(const unsigned *haystack,
 void utf8_wc_strupper(unsigned *str);
 void utf8_wc_strlower(unsigned *str);
 
+int utf8_wc_is_alnum(unsigned wc);
+int utf8_wc_is_space(unsigned wc);
+
+int utf8_table_check(unsigned wc, unsigned const *start, int const *count,
+		     size_t size);
+
 #define DICO_LEV_NORM    0x1
 #define DICO_LEV_DAMERAU 0x2
 
