@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico
-   Copyright (C) 2003-2004, 2007-2008, 2010, 2012 Sergey Poznyakoff
+   Copyright (C) 2003-2018 Sergey Poznyakoff
   
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -233,3 +233,11 @@ dico_assoc_iterator(dico_assoc_list_t assoc)
 	return NULL;
     return dico_list_iterator(assoc->list);
 }
+
+size_t
+dico_assoc_count(dico_assoc_list_t assoc)
+{
+    if (!assoc)
+	return 0;
+    return dico_list_count(assoc->list);
+}
