@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 2008, 2009, 2010 Sergey Poznyakoff
+   Copyright (C) 2008-2018 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -22,12 +22,21 @@ static int sort_index;
 static int trim_ws;
 static int show_dictorg_entries;
 
-static void
-memerr(const char *fname)
-{
-    dico_log(L_ERR, 0, _("%s: not enough memory"), fname);
-}
+#if __STDC_VERSION__ < 199901L
+# if __GNUC__ >= 2
+#  define __func__ __FUNCTION__
+# else
+#  define __func__ "<unknown>"
+# endif
+#endif
 
+#define MEMERR() \
+    dico_log(L_ERR, 0, _("%s:%d:%s: not enough memory"), \
+	     __FILE__, __LINE__, __func__)
+
+typedef int (*COMPARATOR) (const void *, const void *);
+static int get_db_flag(struct dictdb *db, const char *name);
+    
 static int register_strategies(void);
 
 static struct dico_option init_option[] = {
@@ -69,6 +78,13 @@ mod_init(int argc, char **argv)
     return 0;
 }
 
+static void
+dealloc_index_entry(struct index_entry *ent)
+{
+    free(ent->word);
+    free(ent->orig);
+}
+
 static void
 free_db(struct dictdb *db)
 {
@@ -79,7 +95,7 @@ free_db(struct dictdb *db)
     for (i = 0; i < db->numwords; i++) {
 	if (!db->index[i].word)
 	    break;
-	free(db->index[i].word);
+	dealloc_index_entry(&db->index[i]);
     }
     if (db->suf_index) {
 	for (i = 0; i < db->numwords; i++) {
@@ -128,6 +144,7 @@ b64_decode(const char *val, size_t len, size_t *presult)
     return 0;
 }
 
+static COMPARATOR comparator(struct dictdb *db);
 
 static int
 parse_index_entry(const char *filename, size_t line,
@@ -139,12 +156,24 @@ parse_index_entry(const char *filename, size_t line,
     int rc;
     
     memset(&itr, 0, sizeof(itr));
+    memset(&idx, 0, sizeof(idx));
     
     utf8_iter_first(&itr, buf);
 
     rc = 0;
-    for (nfield = 0; nfield < 3; nfield++) {
-	char *start, *end;
+    /* A valid index file contains three to four columns per line:
+         0 - Headword.
+	 1 - Offset of the article in the dictionary file.
+	 2 - Size of the article in bytes
+	 3 - Optional original headword.
+
+      Column 0 is used for searches. Column 3, if present, is used when
+      returning the results of the search. If it is not supplied, column 0
+      is used.
+    */
+    for (nfield = 0; nfield < 4; nfield++) {
+	char const *start;
+	char const *end;
 	size_t len;
 	
 	/* Skip whitespace */
@@ -170,15 +199,25 @@ parse_index_entry(const char *filename, size_t line,
 		while (len > 0 && start[len-1] == ' ')
 		    --len;
 	    }
+
 	    idx.word = malloc(len + 1);
 	    if (!idx.word) {
-		memerr("parse_index_entry");
+		MEMERR();
 		return 1;
 	    }
+
 	    memcpy(idx.word, start, len);
 	    idx.word[len] = 0;
 	    idx.length = len;
 	    idx.wordlen = utf8_strlen(idx.word);
+	} else if (nfield == 3) {
+	    idx.orig = malloc(len + 1);
+	    if (!idx.orig) {
+		MEMERR();
+		return 1;
+	    }
+	    memcpy(idx.orig, start, len);
+	    idx.orig[len] = 0;	    
 	} else {
 	    size_t n;
 	    
@@ -204,7 +243,7 @@ parse_index_entry(const char *filename, size_t line,
 	} else {
 	    ep = malloc(sizeof(*ep));
 	    if (!ep) {
-		memerr("parse_index_entry");
+		MEMERR();
 		rc = 1;
 	    }
 	}
@@ -223,7 +262,7 @@ static int
 free_index_entry(void *item, void *data)
 {
     struct index_entry *ep = item;
-    free(ep->word);
+    dealloc_index_entry(ep);
     free(ep);
     return 0;
 }
@@ -263,7 +302,7 @@ read_index(struct dictdb *db, const char *idxname, int tws)
 
     list = dico_list_create();
     if (!list) {
-	memerr("read_index");
+	MEMERR();
 	rc = 1;
     } else {
 	dico_iterator_t itr;
@@ -313,7 +352,7 @@ open_index(struct dictdb *db, int tws)
     int rc;
     
     if (!idxname) {
-	memerr("open_index");
+	MEMERR();
 	return 1;
     }
 
@@ -370,8 +409,6 @@ open_stream(struct dictdb *db)
     return 1;
 }
 
-static int compare_entry(const void *a, const void *b);
-
 static dico_handle_t
 mod_init_db(const char *dbname, int argc, char **argv)
 {
@@ -413,35 +450,49 @@ mod_init_db(const char *dbname, int argc, char **argv)
 	filename = strdup(filename);
 
     if (!filename) {
-	memerr("mod_init_db");
+	MEMERR();
 	return NULL;
     }
     
-    db = malloc(sizeof(*db));
+    db = calloc(1, sizeof(*db));
     if (!db) {
 	free(filename);
-	memerr("mod_init_db");
+	MEMERR();
 	return NULL;
     }
-    memset(db, 0, sizeof(*db));
 
     db->dbname = dbname;
     db->basename = filename;
     db->show_dictorg_entries = show_dictorg_option;
+    db->flag_allchars = 1;
     
     if (open_index(db, trimws_option)) {
 	free_db(db);
 	return NULL;
     }
 
-    if (sort_option)
-	qsort(db->index, db->numwords, sizeof(db->index[0]), compare_entry);
-    
     if (open_stream(db)) {
 	free_db(db);
 	return NULL;
     }
     
+    db->flag_allchars = get_db_flag(db, DICTORG_FLAG_ALLCHARS);
+    db->flag_casesensitive = get_db_flag(db, DICTORG_FLAG_CASESENSITIVE);
+    db->flag_utf8 = get_db_flag(db, DICTORG_FLAG_UTF8);
+    db->flag_8bit = get_db_flag(db, DICTORG_FLAG_8BIT_NEW);
+    if (get_db_flag(db, DICTORG_FLAG_8BIT_OLD)) {
+	dico_log(L_ERR, 0,
+		 _("mod_init_db(%s): index files in old 8-bit format are not supported"),
+		 argv[0]);
+	free_db(db);
+	return NULL;
+    }
+
+    if (sort_option) {
+	/* Sort index entries */
+	qsort(db->index, db->numwords, sizeof(db->index[0]), comparator(db));
+    }
+    
     return (dico_handle_t)db;
 }
 
@@ -483,8 +534,8 @@ init_suffix_index(struct dictdb *db)
 	for (i = 0; i < db->numwords; i++) {
 	    char *p = malloc(db->index[i].length + 1);
 	    if (!p) {
-		while (i-- >= 0)
-		    free(db->suf_index[i].word);
+		while (i > 0)
+		    free(db->suf_index[--i].word);
 		free(db->suf_index);
 		return 1;
 	    }
@@ -503,11 +554,14 @@ static int exact_match(struct dictdb *, const char *, struct result *);
 static int prefix_match(struct dictdb *, const char *, struct result *);
 static int suffix_match(struct dictdb *, const char *, struct result *);
 
-#define RESERVED_WORD(db, word)			            \
-    (!(db)->show_dictorg_entries                            \
-	&& strlen(word) >= sizeof(DICTORG_ENTRY_PREFIX)-1   \
-	&& memcmp(word, DICTORG_ENTRY_PREFIX,               \
-		  sizeof(DICTORG_ENTRY_PREFIX)-1) == 0)
+#define RESERVED_WORD(db, word)					    \
+    (!(db)->show_dictorg_entries				    \
+     && ((strlen(word) >= DICTORG_ENTRY_PREFIX_LEN		    \
+	 && memcmp(word, DICTORG_ENTRY_PREFIX,			    \
+		  DICTORG_ENTRY_PREFIX_LEN) == 0)		    \
+	|| (strlen(word) >= DICTORG_ALT_ENTRY_PREFIX_LEN	    \
+	&& memcmp(word, DICTORG_ALT_ENTRY_PREFIX,		    \
+		      DICTORG_ALT_ENTRY_PREFIX_LEN) == 0)))	    \
 
 
 static struct strategy_def strat_tab[] = {
@@ -536,7 +590,16 @@ register_strategies(void)
 }
 
 static int
-compare_entry(const void *a, const void *b)
+compare_allchars(const void *a, const void *b)
+{
+    const struct index_entry *epa = a;
+    const struct index_entry *epb = b;
+    compare_count++;
+    return utf8_strcmp(epa->word, epb->word);
+}
+
+static int
+compare_allchars_ci(const void *a, const void *b)
 {
     const struct index_entry *epa = a;
     const struct index_entry *epb = b;
@@ -544,17 +607,52 @@ compare_entry(const void *a, const void *b)
     return utf8_strcasecmp(epa->word, epb->word);
 }
 
+static int
+compare_alnumspace(const void *a, const void *b)
+{
+    const struct index_entry *epa = a;
+    const struct index_entry *epb = b;
+    compare_count++;
+    return utf8_strcmp_alnumspace(epa->word, epb->word);
+}
+
+static int
+compare_alnumspace_ci(const void *a, const void *b)
+{
+    const struct index_entry *epa = a;
+    const struct index_entry *epb = b;
+    compare_count++;
+    return utf8_strcasecmp_alnumspace(epa->word, epb->word);
+}
+
+static COMPARATOR
+comparator(struct dictdb *db)
+{
+    if (db->flag_allchars) {
+	if (db->flag_casesensitive)
+	    return compare_allchars;
+	else
+	    return compare_allchars_ci;
+    } else {
+	if (db->flag_casesensitive)
+	    return compare_alnumspace;
+	else
+	    return compare_alnumspace_ci;
+    }
+}
+
+//FIXME
 static int
 compare_entry_ptr(const void *a, const void *b)
 {
     const struct index_entry *epa = *(const struct index_entry **)a;
     const struct index_entry *epb = *(const struct index_entry **)b;
-    return compare_entry(epa, epb);
+    return compare_alnumspace_ci(epa, epb);
 }
 
 static int
 common_match(struct dictdb *db, const char *word,
-	     int (*compare) (const void *, const void *),
+	     COMPARATOR compare,
 	     int unique, struct result *res)
 {
     struct index_entry x, *ep;
@@ -576,14 +674,14 @@ common_match(struct dictdb *db, const char *word,
 	res->db = db;
 	res->list = dico_list_create();
 	if (!res->list) {
-	    memerr("common_match");
+	    MEMERR();
 	    return 0;
 	}
 	res->itr = NULL;
 	if (unique) {
 	    dico_list_set_comparator(res->list,
 				     (int (*)(const void *, void *))
-				          compare_entry);
+				     comparator(db));
 	    dico_list_set_flags(res->list, DICO_LIST_COMPARE_TAIL);
 	}
 	for (p++; p < ep; p++) 
@@ -599,7 +697,7 @@ common_match(struct dictdb *db, const char *word,
 static int
 exact_match(struct dictdb *db, const char *word, struct result *res)
 {
-    return common_match(db, word, compare_entry, 1, res);
+    return common_match(db, word, comparator(db), 1, res);
 }
 
 static int
@@ -640,14 +738,14 @@ suffix_match(struct dictdb *db, const char *word, struct result *res)
     int rc;
     
     if (init_suffix_index(db)) {
-	memerr("suffix_match");
+	MEMERR();
 	return 1;
     }
     
     ent.length = strlen(word);
     x.word = malloc(ent.length + 1);
     if (!x.word) {
-	memerr("suffix_match");
+	MEMERR();
 	return 1;
     }
     ent.wordlen = utf8_strlen(word);
@@ -677,7 +775,7 @@ suffix_match(struct dictdb *db, const char *word, struct result *res)
 
 	tmp = calloc(count, sizeof(*tmp));
 	if (!tmp) {
-	    memerr("suffix_match");
+	    MEMERR();
 	    free(x.word);
 	    return 1;
 	} 
@@ -691,14 +789,14 @@ suffix_match(struct dictdb *db, const char *word, struct result *res)
 
 	list = dico_list_create();
 	if (!list) {
-	    memerr("suffix_match");
+	    MEMERR();
 	    free(x.word);
 	    free(tmp);
 	    return 1;
 	}
 	dico_list_set_comparator(list,
 				 (int (*)(const void *, void *))
-				   compare_entry);
+				   comparator(db));
 	dico_list_set_flags(list, DICO_LIST_COMPARE_TAIL);
 	for (i = 0; i < count; i++) 
 	    dico_list_append(list, tmp[i]);
@@ -728,12 +826,12 @@ find_db_entry(struct dictdb *db, const char *name)
     x.length = strlen(name);
     x.wordlen = utf8_strlen(name);
     ep = bsearch(&x, db->index, db->numwords, sizeof(db->index[0]),
-		 compare_entry);
+		 comparator(db));
     if (!ep)
 	return NULL;
     buf = malloc(ep->size + 1);
     if (!buf) {
-	memerr("find_db_entry");
+	MEMERR();
 	return NULL;
     }
     dico_stream_seek(db->stream, ep->offset, DICO_SEEK_SET);
@@ -748,6 +846,18 @@ find_db_entry(struct dictdb *db, const char *name)
     return buf;
 }
 
+static int
+get_db_flag(struct dictdb *db, const char *name)
+{
+    struct index_entry x;
+    
+    x.word = (char*) name;
+    x.length = strlen(name);
+    x.wordlen = utf8_strlen(name);
+    return bsearch(&x, db->index, db->numwords, sizeof(db->index[0]),
+		   comparator(db)) != NULL;
+}
+
 static char *
 mod_info(dico_handle_t hp)
 {
@@ -777,15 +887,18 @@ mod_descr(dico_handle_t hp)
     return ptr;
 }
 
+static char *
+mod_mime_header(dico_handle_t hp)
+{
+    struct dictdb *db = (struct dictdb *) hp;
+    return find_db_entry(db, DICTORG_ENTRY_MIME_HEADER);
+}
+
 
 static dico_result_t
-_match_simple(struct dictdb *db, const char *strat, const char *word)
+_match_simple(struct dictdb *db, entry_match_t match, const char *word)
 {
     struct result *res;
-    entry_match_t match = find_matcher(strat);
-
-    if (!match)
-	return NULL;
     
     res = malloc(sizeof(*res));
     if (!res)
@@ -810,13 +923,13 @@ _match_all(struct dictdb *db, dico_strategy_t strat, const char *word)
     list = dico_list_create();
 
     if (!list) {
-	memerr("_match_all");
+	MEMERR();
 	return NULL;
     }
 
     dico_list_set_comparator(list,
 			     (int (*)(const void *, void *))
-			       compare_entry);
+			       comparator(db));
     dico_list_set_flags(list, DICO_LIST_COMPARE_TAIL);
 
     if (dico_key_init(&key, strat, word)) {
@@ -854,14 +967,17 @@ static dico_result_t
 mod_match(dico_handle_t hp, const dico_strategy_t strat, const char *word)
 {
     struct dictdb *db = (struct dictdb *) hp;
+    entry_match_t match;
 
     if (RESERVED_WORD(db, word))
 	return NULL;
-    
-    if (strat->sel) 
+
+    match = find_matcher(strat->name);
+    if (match)
+	return _match_simple(db, match, word);
+    else if (strat->sel) 
 	return _match_all(db, strat, word);
-    else
-	return _match_simple(db, strat->name, word);
+    return NULL;
 }
 
 static dico_result_t
@@ -873,12 +989,13 @@ mod_define(dico_handle_t hp, const char *word)
 
     if (RESERVED_WORD(db, word))
 	return NULL;
-    rc = common_match(db, word, compare_entry, 0, &res);
+    
+    rc = common_match(db, word, comparator(db), 0, &res);
     if (rc)
 	return NULL;
     rp = malloc(sizeof(*rp));
     if (!rp) {
-	memerr("mod_define");
+	MEMERR();
 	dico_list_destroy(&res.list);
 	return NULL;
     }
@@ -929,9 +1046,11 @@ mod_output_result (dico_result_t rp, size_t n, dico_stream_t str)
     
     
     switch (res->type) {
-    case result_match:
-	dico_stream_write(str, ep->word, strlen(ep->word));
+    case result_match: {
+	char *headword = ep->orig ? ep->orig : ep->word;
+	dico_stream_write(str, headword, strlen(headword));
 	break;
+    }
 	
     case result_define:
 	printdef(str, res->db, ep);
@@ -963,20 +1082,18 @@ mod_free_result(dico_result_t rp)
 }
 
 struct dico_database_module DICO_EXPORT(dictorg, module) = {
-    DICO_MODULE_VERSION,
-    DICO_CAPA_NONE,
-    mod_init,
-    mod_init_db,
-    mod_free_db,
-    NULL,
-    NULL,
-    mod_info,
-    mod_descr,
-    NULL, /* FIXME: dico_db_lang */
-    mod_match,
-    mod_define,
-    mod_output_result,
-    mod_result_count,
-    mod_compare_count,
-    mod_free_result
+    .dico_version = DICO_MODULE_VERSION,
+    .dico_capabilities = DICO_CAPA_NONE,
+    .dico_init = mod_init,
+    .dico_init_db = mod_init_db,
+    .dico_free_db = mod_free_db,
+    .dico_db_info = mod_info,
+    .dico_db_descr = mod_descr,
+    .dico_match = mod_match,
+    .dico_define = mod_define,
+    .dico_output_result = mod_output_result,
+    .dico_result_count = mod_result_count,
+    .dico_compare_count = mod_compare_count,
+    .dico_free_result = mod_free_result,
+    .dico_db_mime_header = mod_mime_header
 };
